/*
 * Copyright (C) @BABOLO  2008 May 25
 * Copyright (c) 1985, 1988, 1990, 1992, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
#if 0
static char copyright[] =
"@(#) Copyright (c) 1985, 1988, 1990, 1992, 1993, 1994\n\
	The Regents of the University of California.  All rights reserved.\n";
static char sccsid[] = "@(#)ftpd.c	8.4 (Berkeley) 4/16/94";
static const char rcsid[] = "$FreeBSD: src/libexec/ftpd/ftpd.c,v 1.62.2.61 2005/01/18 21:56:30 yar Exp $";
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 Apr 05\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: wwwftpd.c,v 1.1 2008/06/24 07:05:18 babolo Exp $";
#endif /* 0 */
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 May 25\n\
@(#)Copyright (c) 1985, 1988, 1990, 1992, 1993, 1994\n\
@(#) The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] = "$Id: wwwftpd.c,v 1.1 2008/06/24 07:05:18 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#ifdef	LOGIN_CAP
#include <login_cap.h>
#endif
#include <sys/stat.h>
#include <sysexits.h>
#include <libutil.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <fts.h>
#include <pwd.h>
#include <grp.h>
#include <err.h>
#include <getCGIparm.h>
#include <babolo/BLINflag.h>
#include "wwwftpd.h"

struct	passwd *pw;
char	*homedir;
int	ftpdebug;
int	logging;
int	guest;
int	dochroot;

static volatile sig_atomic_t recvurg;
sig_atomic_t transflag;
off_t	file_size;
off_t	byte_count;
#if !defined(CMASK) || CMASK == 0
#undef CMASK
#define CMASK 027
#endif
int	defumask = CMASK;		/* default umask value */
char	tmpline[7];
char	*hostname;
int	epsvall = 0;

/*
 * Limit number of pathnames that glob can return.
 * A limit of 0 indicates the number of pathnames is unlimited.
 */
#define MAXGLOBARGS	16384
#

#define LOGCMD(cmd, file)                                                                             \
    if  (logging > 1) warnx("INFO %s %s%s", cmd, *(file) == '/' ? "" : curdir(), file);
#define LOGCMD2(cmd, file1, file2)                                                                    \
    if  (logging > 1) warnx( "INFO %s %s%s %s%s", cmd                                                 \
                           , *(file1) == '/' ? "" : curdir(), file1                                   \
                           , *(file2) == '/' ? "" : curdir(), file2                                   \
                           );
#define LOGBYTES(cmd, file, cnt)                                                                      \
    if  (logging > 1) {                                                                               \
        if  (cnt == (off_t)-1) warnx("INFO %s %s%s", cmd, *(file) == '/' ? "" : curdir(), file);      \
          else warnx("INFO %s %s%s = %qd bytes", cmd, (*(file) == '/') ? "" : curdir(), file, cnt);   \
    }
#define ERROUT(A,B) {ex = (A); errno = (B); goto out;}

/*
 * Format and send reply containing system error number.
 */
void
perror_reply(int code, char *string) {
    warnx("%s: %s.", string, strerror(errno));
}

static char *
curdir() {
    static char path[MAXPATHLEN+1+1];	/* path + '/' + '\0' */

    if  (getcwd(path, sizeof(path)-2) == NULL) return ("");
    if  (path[1] != '\0') strcat(path, "/");                    /* special case for root dir. */
    return(guest ? path + 1 : path);         /* For guest account, skip / since it's chrooted */
}

/*
 * Terminate login as previous user, if any, resetting state;
 * used when USER command is given or login fails.
 */
static void
end_login() {
    (void) seteuid((uid_t)0);
    pw = NULL;
#ifdef	LOGIN_CAP
    setusercontext( NULL, getpwuid(0), (uid_t)0
                  , LOGIN_SETPRIORITY | LOGIN_SETRESOURCES | LOGIN_SETUMASK
                  );
#endif
    guest = 0;
    dochroot = 0;
}

/*
 * Record logout in wtmp file
 * and exit with supplied status.
 */
void
dologout(int status) {
    /*
     * Prevent reception of SIGURG from resulting in a resumption
     * back to the main program loop.
     */
    transflag = 0;

    /* beware of flushing buffers after a SIGPIPE */
    _exit(status);
}

void
fatalerror(char *s) {
    warnx("Error in server: %s", s);
    warnx("Closing connection due to server error.");
    dologout(451);
    /* NOTREACHED */
}

/*
 * Helper function for sgetpwnam().
 */
static char *
sgetsave(char *s) {
    char *new = malloc((unsigned) strlen(s) + 1);

    if  (new == NULL) {
        perror_reply(421, "Local resource failure: malloc");
        dologout(1);
        /* NOTREACHED */
    }
    (void) strcpy(new, s);
    return (new);
}

/*
 * Save the result of a getpwnam.  Used for USER command, since
 * the data returned must not be clobbered by any other command
 * (e.g., globbing).
 */
static struct passwd *
sgetpwnam(char *name) {
    static struct passwd save;
    struct passwd *p;

    if  ((p = getpwnam(name)) == NULL) return (p);
    if  (save.pw_name) {
        free(save.pw_name);
        free(save.pw_passwd);
        free(save.pw_gecos);
        free(save.pw_dir);
        free(save.pw_shell);
    }
    save = *p;
    save.pw_name = sgetsave(p->pw_name);
    save.pw_passwd = sgetsave(p->pw_passwd);
    save.pw_gecos = sgetsave(p->pw_gecos);
    save.pw_dir = sgetsave(p->pw_dir);
    save.pw_shell = sgetsave(p->pw_shell);
    return (&save);
}

/*
 * Check if a user is in the file "fname",
 * return a pointer to a malloc'd string with the rest
 * of the matching line in "residue" if not NULL.
 */
static int
checkuser(char *fname, char *name, int pwset, char **residue) {
	FILE *fd;
	int found = 0;
	size_t len;
	char *line, *mp, *p;

	if ((fd = fopen(fname, "r")) != NULL) {
		while (!found && (line = fgetln(fd, &len)) != NULL) {
			/* skip comments */
			if (line[0] == '#')
				continue;
			if (line[len - 1] == '\n') {
				line[len - 1] = '\0';
				mp = NULL;
			} else {
				if ((mp = malloc(len + 1)) == NULL)
					fatalerror("Ran out of memory.");
				memcpy(mp, line, len);
				mp[len] = '\0';
				line = mp;
			}
			/* avoid possible leading and trailing whitespace */
			p = strtok(line, " \t");
			/* skip empty lines */
			if (p == NULL)
				goto nextline;
			/*
			 * if first chr is '@', check group membership
			 */
			if (p[0] == '@') {
				int i = 0;
				struct group *grp;

				if (p[1] == '\0') /* single @ matches anyone */
					found = 1;
				else {
					if ((grp = getgrnam(p+1)) == NULL)
						goto nextline;
					/*
					 * Check user's default group
					 */
					if (pwset && grp->gr_gid == pw->pw_gid)
						found = 1;
					/*
					 * Check supplementary groups
					 */
					while (!found && grp->gr_mem[i])
						found = strcmp(name,
							grp->gr_mem[i++])
							== 0;
				}
			}
			/*
			 * Otherwise, just check for username match
			 */
			else
				found = strcmp(p, name) == 0;
			/*
			 * Save the rest of line to "residue" if matched
			 */
			if (found && residue) {
				if ((p = strtok(NULL, "")) != NULL)
					p += strspn(p, " \t");
				if (p && *p) {
				 	if ((*residue = strdup(p)) == NULL)
						fatalerror("Ran out of memory.");
				} else
					*residue = NULL;
			}
nextline:
			if (mp)
				free(mp);
		}
		(void) fclose(fd);
	}
	return (found);
}

static int login_attempts;	/* number of failed login attempts */
static char curname[MAXLOGNAME];	/* current USER name */

/*
 * USER command.
 * Sets global passwd pointer pw if named account exists and is acceptable;
 * sets askpasswd if a PASS command is expected.  If logged in previously,
 * need to reset state.  If name is "ftp" or "anonymous", the name is not in
 * _PATH_FTPUSERS, and ftp account exists, set guest and pw, then just return.
 * If account doesn't exist, ask for passwd anyway.  Otherwise, check user
 * requesting login privileges.  Disallow anyone who does not have a standard
 * shell as returned by getusershell().  Disallow anyone mentioned in the file
 * _PATH_FTPUSERS to allow people such as root and uucp to be avoided.
 */
int
user(char *name) {
    char *cp, *shell;

    guest = 0;
    if  (strcmp(name, "ftp") == 0 || strcmp(name, "anonymous") == 0) {
        if  (  checkuser(_PATH_FTPUSERS, "ftp", 0, NULL)
            || checkuser(_PATH_FTPUSERS, "anonymous", 0, NULL)
            ) {
            warnx("User %s access denied.", name);
        } else if ((pw = sgetpwnam("ftp")) != NULL) {
            guest = 1;
            warnx("Guest login ok, send your email address as password.");
            return(0);
        } else {
            warnx("User %s unknown.", name);
        }
        return(530);
    }
    if  ((pw = sgetpwnam(name))) {
        if  ((shell = pw->pw_shell) == NULL || *shell == 0) shell = _PATH_BSHELL;
        setusershell();
        while ((cp = getusershell()) != NULL) if (strcmp(cp, shell) == 0) break;
        endusershell();

        if  (cp == NULL || checkuser(_PATH_FTPUSERS, name, 1, NULL)) {
            warnx("User %s access denied.", name);
            pw = (struct passwd *) NULL;
            return(530);
    }   }
    if  (logging) strncpy(curname, name, sizeof(curname)-1);
    return(0);
}

int
pass() {
    int rval;
#ifdef	LOGIN_CAP
    login_cap_t *lc = NULL;
#endif
    char *chrootdir;
    char *residue = NULL;

    if  (!guest) {		/* "ftp" is only account allowed no password */
        if  (pw == NULL) rval = 1; else rval = 0;
 		/*
		 * If rval == 1, the user failed the authentication check
		 * above.  If rval == 0, either PAM or local authentication
		 * succeeded.
		 */
        if  (rval) {
            warnx("Login incorrect.");
            pw = NULL;
            return(530);
        }
    }
    login_attempts = 0;		/* this time successful */
    if  (setegid((gid_t)pw->pw_gid) < 0) {
        warnx("Can't set gid.");
        return(550);
    }
    /* May be overridden by login.conf */
    (void) umask(defumask);
#ifdef	LOGIN_CAP
    lc = login_getpwclass(pw);
    setusercontext( lc, pw, (uid_t)0
                  , LOGIN_SETLOGIN
                  | LOGIN_SETGROUP
                  | LOGIN_SETPRIORITY
                  | LOGIN_SETRESOURCES
                  | LOGIN_SETUMASK
                  );
#else
    setlogin(pw->pw_name);
    (void) initgroups(pw->pw_name, pw->pw_gid);
#endif

    dochroot = checkuser(_PATH_FTPCHROOT, pw->pw_name, 1, &residue)
#ifdef	LOGIN_CAP	/* Allow login.conf configuration as well */
		|| login_getcapbool(lc, "ftp-chroot", 0)
#endif
    ;
    chrootdir = NULL;
	/*
	 * For a chrooted local user,
	 * a) see whether ftpchroot(5) specifies a chroot directory,
	 * b) extract the directory pathname from the line,
	 * c) expand it to the absolute pathname if necessary.
	 */
    if  (dochroot && residue &&
	    (chrootdir = strtok(residue, " \t")) != NULL) {
		if (chrootdir[0] != '/')
			asprintf(&chrootdir, "%s/%s", pw->pw_dir, chrootdir);
		else
			chrootdir = strdup(chrootdir); /* so it can be freed */
		if (chrootdir == NULL)
			fatalerror("Ran out of memory.");
    }
    if  (guest || dochroot) {
		/*
		 * If no chroot directory set yet, use the login directory.
		 * Copy it so it can be modified while pw->pw_dir stays intact.
		 */
        if  (chrootdir == NULL &&
		    (chrootdir = strdup(pw->pw_dir)) == NULL)
			fatalerror("Ran out of memory.");
		/*
		 * Check for the "/chroot/./home" syntax,
		 * separate the chroot and home directory pathnames.
		 */
        if  ((homedir = strstr(chrootdir, "/./")) != NULL) {
			*(homedir++) = '\0';	/* wipe '/' */
			homedir++;		/* skip '.' */
			/* so chrootdir can be freed later */
			if ((homedir = strdup(homedir)) == NULL)
				fatalerror("Ran out of memory.");
        } else {
			/*
			 * We MUST do a chdir() after the chroot. Otherwise
			 * the old current directory will be accessible as "."
			 * outside the new root!
			 */
			homedir = "/";
        }
		/*
		 * Finally, do chroot()
		 */
        if  (chroot(chrootdir) < 0) {
            warnx("Can't change root.");
            goto bad;
        }
    } else	/* real user w/o chroot */
		homedir = pw->pw_dir;
	/*
	 * Set euid *before* doing chdir() so
	 * a) the user won't be carried to a directory that he couldn't reach
	 *    on his own due to no permission to upper path components,
	 * b) NFS mounted homedirs w/restrictive permissions will be accessible
	 *    (uid 0 has no root power over NFS if not mapped explicitly.)
	 */
    if  (seteuid((uid_t)pw->pw_uid) < 0) {
        warnx("Can't set uid.");
        goto bad;
    }
    if  (chdir(homedir) < 0) {
        if  (guest || dochroot) {
            warnx("Can't change to base directory.");
            goto bad;
        } else {
            if  (chdir("/") < 0) {
                warnx("Root is inaccessible.");
                goto bad;
            }
            warnx("No directory! Logging in with home=/");
        }
    }
#ifdef	LOGIN_CAP
    login_close(lc);
#endif
    if  (chrootdir) free(chrootdir);
    if  (residue) free(residue);
    return(0);
bad:
	/* Forget all about it... */
#ifdef	LOGIN_CAP
    login_close(lc);
#endif
    if  (chrootdir) free(chrootdir);
    if  (residue) free(residue);
    end_login();
    return(550);
}

int
retrieve(char *name) {
    mife_descriptor *md;
    int fin, dout;
    struct stat st;
    off_t offset;
    size_t size;

    fin = open(name, O_RDONLY);
    st.st_size = 0;
    if  (fin < 0) {
        if  (errno != 0) {
            perror_reply(550, name);
            LOGCMD("get", name);
        }
        return(550);
    }
    byte_count = -1;
    if  (fstat(fin, &st) < 0) {
        perror_reply(550, name);
        close(fin);
        return(550);
    }
    if  (!S_ISREG(st.st_mode)) {
        /*
         * Never sending a raw directory is a workaround
         * for buggy clients that will attempt to RETR
         * a directory before listing it, e.g., Mozilla.
         * Preventing a guest from getting irregular files
         * is a simple security measure.
         */
        if (S_ISDIR(st.st_mode) || guest) {
            warnx("%s: not a plain file.", name);
            close(fin);
            return(550);
        }
    }
    dout = fileno(stdout);
    if (dout < 0) goto done;
    md = mife_opef(MIFE_BUFX | kukMIFE_BUFL(1), fin);
    for (offset = 0; !(md->flag & MIFE_EOFL); offset += 65536) {
        size = mife_read(md, 65536, offset);
        mife_writ(dout, mife_get(md, offset), size);
    }
    close(dout);
done:
    close(fin);
    return(0);
}

int
store(char *name, int mode) {
    mife_descriptor *md;
    int fout, din;
    off_t offset;
    size_t size;

    fout = open(name, mode ? O_WRONLY | O_CREAT | O_APPEND : O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if  (fout < 0) {
        perror_reply(553, name);
        return(553);
    }
    byte_count = -1;
    din = fileno(stdin);
    if (din < 0) goto done;
    md = mife_opef(MIFE_BUFX | kukMIFE_BUFL(1), din);
    for (offset = 0; !(md->flag & MIFE_EOFL); offset += 65536) {
        size = mife_read(md, 65536, offset);
        mife_writ(fout, mife_get(md, offset), size);
    }
done:
    LOGBYTES(mode ? "append" : "put", name, byte_count);
    close(fout);
    return(0);
}

typedef struct pgoblin_dir {
    void      *dirp;
    int           i;
    int           f;
    char       *ent;
    struct stat *sb;
} pgoblin_dir;

static pgoblin_dir*
local_dir(u_int32_t flags, pgoblin_dir *dir, char *path) {
    pgoblin_dir *d = NULL;
    int ex = EX_OK;
    FTSENT *ent;

    if  (!dir && !path) {
        ifBLIN_QV1(flags) warnx("local: one of dir or path should be");
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (!dir) {
        char *p[2] = {NULL, NULL};

        if  (!(d = calloc(1, sizeof(pgoblin_dir)))) {
            ifBLIN_QV1(flags) warnx("local dir: no mem");
            ERROUT(EX_OSERR, ENOMEM);
        }
        d->i = 0;
        d->f = 1;
        p[0] = path;
        if  (!(d->dirp = fts_open(p, FTS_PHYSICAL | FTS_NOCHDIR | FTS_SEEDOT, NULL))) {
            ifBLIN_QV1(flags) warnx("local dir: no fts");
            ERROUT(EX_OSERR, EIO);
        }
    } else {
        d = dir;
        d->i++;
    }
    for (;;) {
        if  (!(ent = fts_read(d->dirp))) {
            if  (d->dirp) fts_close(d->dirp);
            free(d);
            d = NULL;
            break;
        }
        switch(ent->fts_info) {
        case FTS_SLNONE:
        case FTS_DP:
            continue;
        case FTS_SL:
            fts_set(d->dirp, ent, FTS_FOLLOW);
            continue;
        case FTS_DOT:
        case FTS_DC:
        case FTS_D:
            if  (d->f) {
                d->f = 0;
                continue;
            } else fts_set(d->dirp, ent, FTS_SKIP);
            /* FALLTHRU */
        case FTS_DEFAULT:
        case FTS_F:
            break;
        case FTS_DNR:
            ifBLIN_QV2(flags) warnc(ent->fts_errno, "local dir: fts_read DNR");
            break;
        case FTS_NS:
            ifBLIN_QV2(flags) warnc(ent->fts_errno, "local dir: fts_read NS");
            /* FALLTHRU */
        case FTS_NSOK:
            ent->fts_statp = NULL;
            break;
        case FTS_ERR:
            errno = ent->fts_errno;
            ifBLIN_QV1(flags) warn("local dir: fts_read");
            ex = EX_OSERR;
            goto out;
        default:
            ifBLIN_QV1(flags) warn("local dir: unknown fts_read");
            ERROUT(EX_OSERR, EIO);
        }
        d->sb = ent->fts_statp;
        d->ent = ent->fts_name;
        break;
    }
out:
    if  (ex) {
        if  (d) {
            if  (d->dirp) fts_close(d->dirp);
            free(d);
    }   }
    return(d);
}

int
send_file_list(char *name) {
    pgoblin_dir *dir;
    u_int32_t flags = BLIN_VER1;
    FILE *fi = NULL;
    int ex = EX_OK;
    char *q;

    if  (!(fi = funopen( (const void *)(int)fileno(stdout)
                       , NULL, (int (*)(void*, const char*, int))mife_writ, NULL, NULL
        ) )            ) {
        ifBLIN_QV1(flags) warn("funopen #1");
        ex = EX_OSERR;
    } else {
        for (dir = NULL
            ; (dir = local_dir(flags, dir, name))
            ;) {
            q = dir->ent;
            if  (getCGIparmpass3(0, (u_char*)"copy", (u_char**)&q)) {
                ifBLIN_QV1(flags) warnx("getCGIparmpass3");
                ex = EX_SOFTWARE;
                goto out;
            }
            if  (!(dir->sb)) fprintf(fi, "%d\t%s\t\\N\t\\N\n", dir->i, q);
              else fprintf(fi, "%d\t%s\t%lld\t%u\n", dir->i, q, dir->sb->st_size, dir->sb->st_mode);
            if  (q != dir->ent) free(q);
            q = NULL;
        }
        fprintf(fi, "\\.\n");
    }
out:
    if  (fi && fclose(fi) < 0) {
        ifBLIN_QV1(flags) warn("fclose #1");
        ex = EX_OSERR;
    }
    return(ex);
}

int
delete(char *name) {
    struct stat st;

    LOGCMD("delete", name);
    if  (lstat(name, &st) < 0) {
        perror_reply(550, name);
        return(550);
    }
    if  ((st.st_mode&S_IFMT) == S_IFDIR) {
        if  (rmdir(name) < 0) {
            perror_reply(550, name);
            return(550);
        }
        goto done;
    }
    if  (unlink(name) < 0) {
        perror_reply(550, name);
        return(550);
    }
done:
    return(0);
}

int
renamecmd(char *from, char *to) {
    struct stat st;

    LOGCMD2("rename", from, to);
    if  (guest && (stat(to, &st) == 0)) {
        warnx("%s: permission denied", to);
        return(550);
    }
    if  (rename(from, to) < 0) {
        perror_reply(550, "rename");
        return(550);
    }
    return(0);
}

int
makedir(char *name) {
    if  (mkdir(name, 0777) < 0) {
        perror_reply(550, name);
        return(550);
    }
    return(0);
}

int
removedir(char *name) {
    if  (rmdir(name) < 0) {
        perror_reply(550, name);
        return(550);
    }
    return(0);
}

int
pwd(char *name) {
    char path[MAXPATHLEN + 1], *q, *p;
    int ex = EX_OK;

    if  (realpath(name, path) == NULL) {
        warn("realpath %s", name);
        ex = EX_OSERR;
        goto out;
    }
    if  (path[0] != '/') {
        warnx("realpath %s %s", name, path);
        ex = EX_OSERR;
        goto out;
    }
    for (q = &path[1]; *q; q = p + 1) {
        if  ((p = index(q, '/'))) *p = '\0';
        printf("%s\t%s\n", q, path);
        if  (!p) break;
        *p = '/';
    }
out:
    return(ex);
}

static void
sigurg(int signo) {
    recvurg = 1;
}

static void
sigquit(int signo) {
    warnx("ERR got signal %d", signo);
    dologout(1);
}

static void
lostconn(int signo) {
    if  (ftpdebug) warnx("DEBUG lost connection");
    dologout(1);
}

int
main(int argc, char *argv[], char **envp) {
    char *name = NULL, *cmd = NULL, *parm1 = NULL, *parm2 = NULL;
    struct sigaction sa;
    int ch;

    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;

    while ((ch = getopt(argc, argv, "lu:v")) != -1) {
        switch (ch) {
        case 'l': logging++;	/* > 1 == extra logging */                 break;
        case 'u': {   long val = 0;
                ;
                ;     val = strtol(optarg, &optarg, 8);
                ;     if  (*optarg != '\0' || val < 0) {
                ;         warnx("WARNING bad value for -u");
                ;     } else defumask = val;
                ;     break;
                ; }
        case 'v': ftpdebug++;                                              break;
        default : warnx("WARNING unknown flag -%c ignored", optopt);       break;
        }
    }
    argc -= optind;
    argv += optind;

    if  (argc <= 0 || !(name = argv[0])) {
        warnx("no user name");
        exit(1);
    }
    if  (argc <= 1 || !(cmd = argv[1])) {
        warnx("no command");
        exit(1);
    }
    if  (argc <= 2 || !(parm1 = argv[2]) || !*parm1) {
        warnx("no parameter");
        exit(1);
    }
    if  (argc <= 3) parm1 = argv[2];
    sa.sa_handler = SIG_DFL;
    (void)sigaction(SIGCHLD, &sa, NULL);

    sa.sa_handler = sigurg;
    sa.sa_flags = 0;                      /* don't restart syscalls for SIGURG */
    (void)sigaction(SIGURG, &sa, NULL);
    sigfillset(&sa.sa_mask);                   /* block all signals in handler */
    sa.sa_flags = SA_RESTART;
    sa.sa_handler = sigquit;
    (void)sigaction(SIGHUP, &sa, NULL);
    (void)sigaction(SIGINT, &sa, NULL);
    (void)sigaction(SIGQUIT, &sa, NULL);
    (void)sigaction(SIGTERM, &sa, NULL);
    sa.sa_handler = lostconn;
    (void)sigaction(SIGPIPE, &sa, NULL);
    tmpline[0] = '\0';
    if  ((ch = user(name))) exit(ch);
    if  ((ch = pass())) exit(ch);
    if  (!strcmp(cmd, "RETR")) {
        ch = retrieve(parm1);
    } else if (!strcmp(cmd, "STOR")) {
        ch = store(parm1, 0);
    } else if (!strcmp(cmd, "APPE")) {
        ch = store(parm1, 1);
    } else if (!strcmp(cmd, "NLST")) {
        ch = send_file_list(parm1);
    } else if (!strcmp(cmd, "DELE")) {
        ch = delete(parm1);
    } else if (!strcmp(cmd, "RNTO")) {
        ch = renamecmd(parm1, parm2);
    } else if (!strcmp(cmd, "MKD")) {
        ch = makedir(parm1);
    } else if (!strcmp(cmd, "RMD")) {
        ch = removedir(parm1);
    } else if (!strcmp(cmd, "PWD")) {
        ch = pwd(parm1);
    } else {
        ch = 1;
        warnx("unknown command");
    }
    exit(ch);
}
